#pragma once

/*****************************************************************

アプリケーション固有のプロパティーを管理する.

ChangePropertyでプロパティー値を受け取り,

INotifyProperty* m_reciever;
として登録されているレシーバーに通知(Notify)する.

RegisterNotifyReciever()によってm_recieverにレシーバーを登録する.

******************************************************************/


#include <cstdlib>
#include "INotifyProperty.h"
#include "PropertyList.h"

#include "RenderingProperty.h"


class AIScopeProperty
{
private:
	INotifyProperty* m_reciever;

	//property一覧//
	RenderingProperty m_rendering_property;
	
public:
	AIScopeProperty() :
		m_reciever(NULL)
	{
	};


	void RegisterNotifyReciever(INotifyProperty* reciever){
		m_reciever = reciever;
	}

	void ChangeProperty(int key, int value)
	{
		if (mSetPropertyInList(key, value) > 0){
			//成功した場合//
			if (m_reciever)	{
				m_reciever->Notify(key, value);
			}
		}

	}

private:
	int mSetPropertyInList(int key, int value){

		/*
			アプリケーション固有のプロパティーを記述
		*/
		switch (key){
		case AISCUPE_KEY_PROJECTION:
			m_rendering_property.projection_mode = value;
			break;
		case AISCUPE_KEY_BOX_DRAW:
			m_rendering_property.box_draw = value;
			break;
		case AISCUPE_KEY_MULTIVIEW:
			m_rendering_property.multiview_mode = value;
			break;
		case AISCUPE_KEY_ATOM_DRAW:
			m_rendering_property.atom_draw = value;
			break;
		case AISCUPE_KEY_ATOM_COLOR:
			m_rendering_property.atom_color = value;
			break;
		case AISCUPE_KEY_BOND_DRAW:
			m_rendering_property.bond_draw = value;
			break;

		default:
			return -1;
		}

		return key;
	}

public:
	//3Dレンダリング時に使用するプロパティー値を返す//
	void GetRenderingProperty(RenderingProperty* rendering_property){
		if (rendering_property == NULL){
			return;
		}

		*rendering_property = m_rendering_property;

	};

};