#pragma once
#include "SphereRendererDX11.h"
#include "GeometryCylinder.h"


template <typename GEO>
class StripMaker {
public:

	StripMaker(GEO& geometry) :
		m_geometry(geometry),
		is_begin_point(true),
		add_numbber(0)
	{

	}


	void Push(const vec3f& point, const uint32_t color) {

		if (add_numbber == 0) {
			add_numbber++;
		}else if (add_numbber == 1) {
			m_geometry.PushFirst(m_point_org, point,
				m_color_org,
				m_vertexes,
				m_indexes);
			add_numbber++;
		} else {
			m_geometry.PushSecond(m_point_org, point,
				m_color_org,
				m_vertexes,
				m_indexes);
		}

		m_point_org = point;
		m_color_org = color;

	}

	//区切る//
	void Split() {
		if (add_numbber > 1) {
			m_geometry.PushLast(m_point_org,
				m_color_org,
				m_vertexes,
				m_indexes);
		}
		add_numbber = 0;
	}


	void GetSize(size_t* v_size, size_t* i_size) {
		*v_size = m_vertexes.size();
		*i_size = m_indexes.size();
	}


	VERTEX_F3F3UI1* GetVertexBuffer() {
		return &(m_vertexes[0]);
	}

	int* GetIndexBuffer() {
		return &(m_indexes[0]);
	}

	void Clear() {
		is_begin_point = true;
		m_vertexes.clear();
		m_indexes.clear();
		add_numbber = 0;
	}

private:
	GEO& m_geometry;
	vec3f m_point_org;
	uint32_t m_color_org;

	vector<VERTEX_F3F3UI1> m_vertexes;
	vector<int> m_indexes;

	bool is_begin_point;
	int add_numbber;

};


class CylinderStripRendererDX11  {

public:
	CylinderStripRendererDX11(ID3D11Device* pD3DDevice, ID3D11DeviceContext* pD3DDeviceContext, CommonShader* comShader);
	virtual ~CylinderStripRendererDX11();

	void Draw(const vec3f* r, const int* strip_indexes, const int* strip_lengths, const int num_strips, const uint32_t color, const VISUAL_SETTING& vis);

	void SetViewAndProjectionMatrix(const float* pViewMatrix, const float* pProjectionMatrix) {
		renderer.SetViewAndProjectionMatrix(pViewMatrix, pProjectionMatrix);
	}

private:

	RendererWrapper renderer;
	void InitShader();

	int m_vram_size;
	unsigned int m_index_type;

	
	//ポリゴンモデル用のバッファ
	GeometryCylinder* m_geometry_bond;
	StripMaker<GeometryCylinder>* m_maker;

	VISUAL_SETTING m_prev_vis;

	int m_visual_bond;	//ボンドの表示種類
	



};


