// PointRendererDX11.h: PointRendererDX11 クラスのインターフェイス
//
//////////////////////////////////////////////////////////////////////

#ifndef PointRendererDX11_h
#define PointRendererDX11_h

#pragma once

#include <d3d11.h>


#include "MDLoader.h"	//ATOMS_DATAのためだけに必要//
#include "BOND_INFO.h"

//#include "sdkwgl.h"
#include "vec3.h"
#include "mat33.h"
#include "mat44.h"
#include "io64.h"
#include "mod.h"


#include "visual.h"


#include "BasicRenderer.h"
#include "VertexTypes.h"





class BasePointRendererDX11 : public BasicRenderer {

protected:
	ID3D11GeometryShader*     m_pGeometryShader;
	ID3D11Buffer* m_cbPoint;

	StateManagerDX11 m_state_manager;

	struct vec2f {
		float	x;
		float	y;
	} m_pPointSize;


	enum POINT_SPRITE_LAYOUT {
		RGB_FLOAT_COLOR_U8,
		RGB_FLOAT_COLOR_FLOAT
	};

	
	HRESULT InitShader(POINT_SPRITE_LAYOUT layout_mode);

	
	virtual void DrawOnlyVertex(UINT particle_count, D3D_PRIMITIVE_TOPOLOGY primitive_topology);

public:

	BasePointRendererDX11(ID3D11Device* pD3DDevice, ID3D11DeviceContext* pD3DDeviceContext, CommonShader* comShader);
	virtual ~BasePointRendererDX11();



	void SetPointSize(float width) {
		m_pPointSize.x = width;
		m_pPointSize.y = width;
	}


};



class PointRendererDX11 : public BasePointRendererDX11 {
	
private:

	int m_vram_size;
	int m_vbo_vcount;
	VERTEX_F3UI1* m_vbo_vertex;
	
	unsigned int *m_color_table;	//原子ごとの色が入る

	ATOMS_DATA* m_prev_render_dat;	//前回レンダリングしたデータ
	void CreateSphere(int start, int count, ATOMS_DATA* dat, BOND_INFO* bond);
	
	void CheckColorByBondNum( ATOMS_DATA* dat);
	void CheckColorByBondNum2( ATOMS_DATA* dat, BOND_INFO* bond);
	void CheckColorByAtomKind( ATOMS_DATA* dat);
	void CheckColorByPressure( ATOMS_DATA* dat, BOND_INFO* bond );
	
	VISUAL_SETTING m_prev_vis;

public:
	
	PointRendererDX11(ID3D11Device* pD3DDevice, ID3D11DeviceContext* pD3DDeviceContext, CommonShader* comShader);
	virtual ~PointRendererDX11();
	
	void Draw(ATOMS_DATA* dat, BOND_INFO* bond, const VISUAL_SETTING& vis);
		
};





#endif // PointRendererDX11

