
#pragma warning(disable : 4996)

#define WIN32_LEAN_AND_MEAN             // Windows ヘッダーから使用されていない部分を除外します。
#define _USE_MATH_DEFINES

#include <tchar.h>
#include <string.h>
#include <math.h>

#include "RenderDX11.h"

#include "windib2.h"
#include "WriteBMP.h"

#include "mod.h"
#include "visual.h"
#include "camera.h"
#include "atomic_color.h"


#include "SphereRendererDX11.h"
#include "SphereColorTableRendererDX11.h"
#include "PointRendererDX11.h"
#include "PointSpriteDX11.h"
#include "VNTRenderer.h"
#include "KrbRendererDX11.h"
#include "BoxRendererDX11.h"
#include "CylinderStripRendererDX11.h"


#include "MDLoader.h"
#include "filelist.h"


#include "setting.h"


extern double VIEW_ANGLE;
extern double VIEW_NEAR;
extern double VIEW_FAR;



//extern MasterFileList g_files;		//開いたファイルのリスト.
extern bool g_updateFlag;	//レンダリングをしないようにLockする。メモリ転送など.
extern int g_filter_mode;

namespace {
	//DirectX11 system//////////////////////////////
	DX11_CORE s_dx11_core;
	CommonShader* comShader = nullptr;
	
	SphereRendererDX11* m_sphere_renderer = nullptr;
	CylinderRendererDX11* m_cylinder_renderer = nullptr;
	SphereColorTableRendererDX11* m_sphere_color_table_renderer = nullptr;
	PointRendererDX11* m_pointRendererDX11 = nullptr;
	PointSpriteDX11* m_pointSpriteDX11 = nullptr;
	VNTRenderer* m_vntRenderer = nullptr;
	KrbRendererDX11* m_karabaRendererDX11 = nullptr;
	BoxRendererDX11* m_boxRendererDX11 = nullptr;
	CylinderStripRendererDX11* m_strip_renderer = nullptr;

	//polygon//////////////////////////////////////////////
	vec3f* flier_vertex;
	int flier_vertex_count;
	WORD* flier_index;
	int flier_index_count;

	///////////////////////////////////////////////////////////////////
	//変換行列用ライブラリ
	///////////////////////////////////////////////////////////////////
	void SetProjectionPerspectiveMatrix(float angle, float wph, float z_near, float z_far, mat44f* pProjectionMatrix);
	void SetProjectionRange(int projection_mode, int image_w, int image_h, int range_x, int range_y, int range_w, int range_h, double distance, mat44f* projection_matrix);

	StateManagerDX11* state_manager_dx;

	enum FILTER_MODE {
		FILTER_NONE = 0,
		FILTER_DEPTH_OF_FIELD = 1            //ぼかし//
	};
	


	class FilterDepthOfField {
	private:
		DX11_CORE* m_dx11_core;
		D3D11_TEXTURE2D_DESC m_previous_descBackBuffer{ 0 };
		ID3D11RenderTargetView* m_filterRTV;
		ID3D11DepthStencilView* m_filterDSV;
		ID3D11ShaderResourceView* m_renderSRV;
		ID3D11ShaderResourceView* m_depthSRV;
		ID3D11SamplerState*     m_pSampler;
		ID3D11Buffer* m_texel_width_CB;

		RendererWrapper renderer;

	public:
		FilterDepthOfField(DX11_CORE* dx11_core, CommonShader* comShader) :
			renderer(dx11_core->pD3DDevice, dx11_core->pD3DDeviceContext, comShader),
			m_dx11_core(dx11_core),
			m_filterRTV(NULL),
			m_renderSRV(NULL),
			m_filterDSV(NULL),
			m_depthSRV(NULL),
			m_pSampler(NULL),
			m_texel_width_CB(NULL)
		{

			//シェーダー//
			TCHAR filepath[MAX_PATH];
			D3D11_INPUT_ELEMENT_DESC layout[] =
			{
				{ "POSITION", 0, DXGI_FORMAT_R32G32B32_FLOAT, 0, 0, D3D11_INPUT_PER_VERTEX_DATA, 0 },
				{ "TEX", 0, DXGI_FORMAT_R32G32_FLOAT, 0, 12, D3D11_INPUT_PER_VERTEX_DATA, 0 }
			};
			UINT numElements = ARRAYSIZE(layout);
			GetAttachedFileName(filepath, _T("vs_filter_DoF.cso"));
			renderer.LoadVertexShader(filepath, "main", "vs_4_0", layout, numElements);


			//Pixel Shader-----------------------------------------
			GetAttachedFileName(filepath, _T("ps_filter_DoF.cso"));
			renderer.LoadPixelShader(filepath, "main", "ps_4_0");
		

			// テクスチャのサンプリング(ポリゴン貼り付け時の補間)方法の指定.
			D3D11_SAMPLER_DESC sampDesc;
			ZeroMemory(&sampDesc, sizeof(sampDesc));
			sampDesc.Filter = D3D11_FILTER_MIN_MAG_MIP_LINEAR;
			sampDesc.AddressU = D3D11_TEXTURE_ADDRESS_WRAP;
			sampDesc.AddressV = D3D11_TEXTURE_ADDRESS_WRAP;
			sampDesc.AddressW = D3D11_TEXTURE_ADDRESS_WRAP;
			sampDesc.ComparisonFunc = D3D11_COMPARISON_NEVER;
			sampDesc.MinLOD = 0;
			sampDesc.MaxLOD = D3D11_FLOAT32_MAX;
			HRESULT hr = m_dx11_core->pD3DDevice->CreateSamplerState(&sampDesc, &m_pSampler);
			

			//テクセル幅(1/画面幅)を伝えるための定数バッファ//
			D3D11_BUFFER_DESC bd;
			bd.ByteWidth = sizeof(float) * 4;	//16の倍数である必要がある//
			bd.Usage = D3D11_USAGE_DYNAMIC;
			bd.BindFlags = D3D11_BIND_CONSTANT_BUFFER;
			bd.CPUAccessFlags = D3D11_CPU_ACCESS_WRITE;	//CPUからの書き換え
			bd.MiscFlags = 0;
			bd.StructureByteStride = 0;
				
			hr = m_dx11_core->pD3DDevice->CreateBuffer(&bd, NULL, &m_texel_width_CB);
		}
		
		~FilterDepthOfField() {
			SAFE_RELEASE(m_filterRTV);
			SAFE_RELEASE(m_renderSRV);
			SAFE_RELEASE(m_pSampler);
			SAFE_RELEASE(m_texel_width_CB);
			SAFE_RELEASE(m_filterDSV);
			SAFE_RELEASE(m_depthSRV);
		}

	private:
		bool IsEqual(D3D11_TEXTURE2D_DESC a, D3D11_TEXTURE2D_DESC b) {
			if (a.ArraySize != b.ArraySize) return false;
			if (a.Format != b.Format) return false;
			if (a.Height != b.Height) return false;
			if (a.Width != b.Width) return false;
			if (a.MipLevels != b.MipLevels) return false;
			if (a.MiscFlags != b.MiscFlags) return false;
			if (a.SampleDesc.Count != b.SampleDesc.Count) return false;
			if (a.SampleDesc.Quality != b.SampleDesc.Quality) return false;
			return true;
		}

	public:
		void Render(vector<LawData>& data_list, int screen_w, int screen_h, int range_x, int range_y, int range_w, int range_h, const double* view_matrix, const double focus_distance, const RenderingProperty& rendering_property) {
			
			//(2)スクリーンフォーマットを取得//
			ID3D11Texture2D* pBackBuffer = NULL;
			HRESULT hr = m_dx11_core->pSwapChain->GetBuffer(0, __uuidof(ID3D11Texture2D), (LPVOID*)&pBackBuffer);
			if (FAILED(hr)) return;

			D3D11_TEXTURE2D_DESC descBackBuffer;
			pBackBuffer->GetDesc(&descBackBuffer);
			pBackBuffer->Release();

			//サイズが異なったらrendertargetを作成しなおす//
			if (!IsEqual(m_previous_descBackBuffer, descBackBuffer)) {
				m_previous_descBackBuffer = descBackBuffer;

				//オフスクリーンレンダリング用のレンダリングターゲット(texture)バッファをGPU上に作成//
				D3D11_TEXTURE2D_DESC Texture2DDesc;
				Texture2DDesc.ArraySize = descBackBuffer.ArraySize;
				Texture2DDesc.BindFlags = D3D11_BIND_RENDER_TARGET | D3D11_BIND_SHADER_RESOURCE;
				Texture2DDesc.CPUAccessFlags = 0;
				Texture2DDesc.Format = descBackBuffer.Format;
				Texture2DDesc.Height = descBackBuffer.Height;
				Texture2DDesc.Width = descBackBuffer.Width;
				Texture2DDesc.MipLevels = descBackBuffer.MipLevels;
				Texture2DDesc.MiscFlags = descBackBuffer.MiscFlags;
				Texture2DDesc.SampleDesc.Count = descBackBuffer.SampleDesc.Count;
				Texture2DDesc.SampleDesc.Quality = descBackBuffer.SampleDesc.Quality;
				Texture2DDesc.Usage = D3D11_USAGE_DEFAULT;               //GPUで書き込むにはD3D11_USAGE_DEFAULTにする必要がある//

				ID3D11Texture2D *hCaptureTexture;
				hr = m_dx11_core->pD3DDevice->CreateTexture2D(&Texture2DDesc, 0, &hCaptureTexture);

				SAFE_RELEASE(m_filterRTV);
				hr = m_dx11_core->pD3DDevice->CreateRenderTargetView(hCaptureTexture, NULL, &m_filterRTV);
				
				{
					SAFE_RELEASE(m_renderSRV);
					D3D11_SHADER_RESOURCE_VIEW_DESC srv_desc;
					memset(&srv_desc, 0, sizeof(srv_desc));
					srv_desc.Format = descBackBuffer.Format;
					srv_desc.ViewDimension = D3D11_SRV_DIMENSION_TEXTURE2D;
					srv_desc.Texture2D.MipLevels = 1;
					hr = m_dx11_core->pD3DDevice->CreateShaderResourceView(hCaptureTexture, &srv_desc, &m_renderSRV);
				}

				hCaptureTexture->Release();


				//オフスクリーンレンダリング用のレンダリングターゲット(深度/ステンシル)バッファをGPU上に作成//
				D3D11_TEXTURE2D_DESC descDepth;
				descDepth.Width = descBackBuffer.Width;   // 幅
				descDepth.Height = descBackBuffer.Height;  // 高さ
				descDepth.MipLevels = 1;       // ミップマップ・レベル数
				descDepth.ArraySize = 1;       // 配列サイズ
				descDepth.Format = DXGI_FORMAT_R24G8_TYPELESS;  // フォーマットShaderResourceにするために使えるフォーマットは限られる//
				descDepth.SampleDesc.Count = descBackBuffer.SampleDesc.Count;  // マルチサンプリングの設定
				descDepth.SampleDesc.Quality = descBackBuffer.SampleDesc.Quality;  // マルチサンプリングの品質
				descDepth.Usage = D3D11_USAGE_DEFAULT;      // デフォルト使用法
				descDepth.BindFlags = D3D11_BIND_DEPTH_STENCIL | D3D11_BIND_SHADER_RESOURCE; // 深度/ステンシルとして使用
				descDepth.CPUAccessFlags = 0;   // CPUからはアクセスしない
				descDepth.MiscFlags = 0;   // その他の設定なし
				ID3D11Texture2D* filterDepthStencilBuffer;
				//2Dテクスチャの作成//
				hr = m_dx11_core->pD3DDevice->CreateTexture2D(&descDepth, NULL, &filterDepthStencilBuffer);
				
				SAFE_RELEASE(m_filterDSV);
				D3D11_DEPTH_STENCIL_VIEW_DESC descDSV;
				descDSV.Format = DXGI_FORMAT_D24_UNORM_S8_UINT;//バッファ作成時とは異なるフォーマットにする(RenderTargetとして使えるフォーマットの制限のため)
				descDSV.ViewDimension = D3D11_DSV_DIMENSION_TEXTURE2D;	//マルチサンプリングならD3D10_DSV_DIMENSION_TEXTURE2DMS
				descDSV.Flags = 0;		//DX11
				descDSV.Texture2D.MipSlice = 0;		//マルチサンプリングならTexture2DMS
				// 深度/ステンシル・ビューの作成//
				hr = m_dx11_core->pD3DDevice->CreateDepthStencilView(filterDepthStencilBuffer, &descDSV, &m_filterDSV);

				{
					SAFE_RELEASE(m_depthSRV);
					D3D11_SHADER_RESOURCE_VIEW_DESC srv_desc;
					memset(&srv_desc, 0, sizeof(srv_desc));
					srv_desc.Format = DXGI_FORMAT_R24_UNORM_X8_TYPELESS;//バッファ作成時とは異なるフォーマットにする(ShaderResourceとして使えるフォーマットの制限のため)
					srv_desc.ViewDimension = D3D11_SRV_DIMENSION_TEXTURE2D;
					srv_desc.Texture2D.MipLevels = 1;
					hr = m_dx11_core->pD3DDevice->CreateShaderResourceView(filterDepthStencilBuffer, &srv_desc, &m_depthSRV);
				}

				filterDepthStencilBuffer->Release();

			}
			
			//(3.5)レンダリングターゲットの差し替え//
			ID3D11RenderTargetView* org_RTV;
			ID3D11DepthStencilView* org_DSV;
			m_dx11_core->pD3DDeviceContext->OMGetRenderTargets(1, &org_RTV, &org_DSV);
			m_dx11_core->pD3DDeviceContext->OMSetRenderTargets(1, &m_filterRTV, m_filterDSV);
			

			//BackBufferのクリア//
			m_dx11_core->pD3DDeviceContext->ClearRenderTargetView(m_filterRTV, g_bg_color);
			//DepthBufferのクリア//
			m_dx11_core->pD3DDeviceContext->ClearDepthStencilView(m_filterDSV, D3D11_CLEAR_DEPTH, 1.0f, 0);
			RenderingCoreDX11(data_list, screen_w, screen_h, range_x, range_y, range_w, range_h, view_matrix, focus_distance, rendering_property);

			
			//レンダリングターゲットを戻す//
			m_dx11_core->pD3DDeviceContext->OMSetRenderTargets(1, &org_RTV, org_DSV);
			SAFE_RELEASE(org_RTV);
			SAFE_RELEASE(org_DSV);
			
			//ビューポートの設定(RenderingCoreDX11でマルチレンダリングの場合に備えて)//
			D3D11_VIEWPORT vp;
			vp.Width = (FLOAT)range_w;
			vp.Height = (FLOAT)range_h;
			vp.MinDepth = 0.0f;
			vp.MaxDepth = 1.0f;
			vp.TopLeftX = 0;
			vp.TopLeftY = 0;
			m_dx11_core->pD3DDeviceContext->RSSetViewports(1, &vp);
			

			ClearRenderingTargetDX11();
			m_dx11_core->pD3DDeviceContext->PSSetSamplers(0, 1, &m_pSampler);
			
			m_dx11_core->pD3DDeviceContext->PSSetShaderResources(0, 1, &m_renderSRV);
			m_dx11_core->pD3DDeviceContext->PSSetShaderResources(1, 1, &m_depthSRV);
			

			VERTEX_F3F2 vertex[4];
			vertex[0].position.Set(-1.0f, -1.0f, .0f);
			vertex[1].position.Set(-1.0f, 1.0f, .0f);
			vertex[2].position.Set(1.0f, 1.0f, .0f);
			vertex[3].position.Set(1.0f, -1.0f, .0f);
			const float u_max = (float)range_w / (float)descBackBuffer.Width;//画面幅とテクスチャ幅が異なる場合の対策//
			const float v_max = (float)range_h / (float)descBackBuffer.Height;
			vertex[0].coord_u = 0.0f;
			vertex[0].coord_v = v_max;
			vertex[1].coord_u = 0.0f;
			vertex[1].coord_v = 0.0f;
			vertex[2].coord_u = u_max;
			vertex[2].coord_v = 0.0f;
			vertex[3].coord_u = u_max;
			vertex[3].coord_v = v_max;
			
			int indexes[6]{ 0,1,2,2,3,0 };
			
			renderer.UpdateVertexBuffer(vertex, 4, sizeof(VERTEX_F3F2));	//頂点データをVRAMに転送
			renderer.UpdateIndexBuffer(indexes, 6, sizeof(int));	//インデックスデータをVRAMに転送

			//texel幅の送信(//定数バッファの書き込み)//
			{
				float texel_width[2] = {1.0f/ (float)descBackBuffer.Width , 1.0f/ (float)descBackBuffer.Height};				
				D3D11_MAPPED_SUBRESOURCE mappedResource;
				m_dx11_core->pD3DDeviceContext->Map(m_texel_width_CB, 0, D3D11_MAP_WRITE_DISCARD, 0, &mappedResource);
				CopyMemory(mappedResource.pData, &texel_width, sizeof(float)*2);
				m_dx11_core->pD3DDeviceContext->Unmap(m_texel_width_CB, 0);
				m_dx11_core->pD3DDeviceContext->PSSetConstantBuffers(0, 1, &m_texel_width_CB);
			}

			renderer.DrawIndexed(6, D3D_PRIMITIVE_TOPOLOGY_TRIANGLELIST);

		}
	};

	FilterDepthOfField* s_filterDoF;
}


///////////////////////////////////////////////////////////////////
//以下はユーザー定義のコールバック関数
///////////////////////////////////////////////////////////////////

void InitDX11(DX11_CORE& dx11_core){
	
	s_dx11_core = dx11_core;
	
	//Shader共用管理クラス
	comShader = new CommonShader(dx11_core.pD3DDevice);
	
	state_manager_dx = new StateManagerDX11(dx11_core.pD3DDevice, dx11_core.pD3DDeviceContext);

	// ブレンド・ステート・オブジェクトの作成
	// パーティクル用ブレンド・ステート・オブジェクト
	D3D11_BLEND_DESC blendDesc;
	ZeroMemory(&blendDesc, sizeof(D3D11_BLEND_DESC));
	blendDesc.AlphaToCoverageEnable = FALSE;
	blendDesc.IndependentBlendEnable = FALSE;
	blendDesc.RenderTarget[0].BlendEnable = FALSE;
	blendDesc.RenderTarget[0].BlendOp = D3D11_BLEND_OP_ADD;
	blendDesc.RenderTarget[0].SrcBlend = D3D11_BLEND_SRC_ALPHA;//D3D11_BLEND_ONE;
	blendDesc.RenderTarget[0].DestBlend = D3D11_BLEND_ONE;
	blendDesc.RenderTarget[0].BlendOpAlpha = D3D11_BLEND_OP_ADD;
	blendDesc.RenderTarget[0].SrcBlendAlpha = D3D11_BLEND_ZERO;
	blendDesc.RenderTarget[0].DestBlendAlpha = D3D11_BLEND_ZERO;
	blendDesc.RenderTarget[0].RenderTargetWriteMask = D3D11_COLOR_WRITE_ENABLE_ALL;

	//	BlendFunc(&blendDesc, DOG_SRC_ALPHA, DOG_ONE_MINUS_SRC_ALPHA);

	state_manager_dx->SetBlendState(&blendDesc);


	//深度バッファを書き込みしない.
	// デプス・ステート・オブジェクトの作成
	// パーティクル用デプス・ステート・オブジェクト
	//ポイントスプライトではデプスバッファのチェックだけして、更新はしない.
	D3D11_DEPTH_STENCIL_DESC depthStencilState;
	ZeroMemory(&depthStencilState, sizeof(D3D11_DEPTH_STENCIL_DESC));
	depthStencilState.DepthEnable = TRUE;
	depthStencilState.DepthWriteMask = D3D11_DEPTH_WRITE_MASK_ALL;	//深度バッファを書き込みする.
	depthStencilState.DepthFunc = D3D11_COMPARISON_LESS;
	depthStencilState.StencilEnable = FALSE;
	depthStencilState.StencilReadMask = D3D11_DEFAULT_STENCIL_READ_MASK;
	depthStencilState.StencilWriteMask = D3D11_DEFAULT_STENCIL_WRITE_MASK;
	//	depthStencilState.StencilFunc = D3D11_STENCIL_OP_KEEP;
	//	depthStencilState.StencilDepthFailOp = D3D11_STENCIL_OP_KEEP;
	//	depthStencilState.StencilPassOp = D3D11_STENCIL_OP_KEEP;
	//	depthStencilState.StencilFailOp = D3D11_COMPARISON_ALWAYS;
	state_manager_dx->SetDepthStencilState(&depthStencilState);


	// ラスタライザの作成//
	//カリング等を制御//
	D3D11_RASTERIZER_DESC rasterizerState;
	ZeroMemory(&rasterizerState, sizeof(D3D11_RASTERIZER_DESC));
	rasterizerState.FillMode = D3D11_FILL_SOLID;
	rasterizerState.CullMode = D3D11_CULL_BACK;
	rasterizerState.FrontCounterClockwise = TRUE;
	rasterizerState.DepthBias = 0;
	rasterizerState.DepthBiasClamp = 0;
	rasterizerState.SlopeScaledDepthBias = 0;
	rasterizerState.DepthClipEnable = FALSE;
	rasterizerState.ScissorEnable = FALSE;
	rasterizerState.MultisampleEnable = FALSE;
	rasterizerState.AntialiasedLineEnable = FALSE;
	state_manager_dx->SetRasterizerState(&rasterizerState);


	s_filterDoF = new FilterDepthOfField(&s_dx11_core, comShader);

}

void ResetDX11(DX11_CORE& dx11_core){

	s_dx11_core = dx11_core;

}

void TerminateDX11(){

	if(m_pointRendererDX11){ delete m_pointRendererDX11;}
	//if(m_atomRendererDX11){ delete m_atomRendererDX11;}
	
	delete state_manager_dx;
	delete s_filterDoF;

	if(comShader){ delete comShader;}

}


template <typename F, typename D>
static void Transpose(F* matrix44, const D* source44){
	matrix44[0] = (F)source44[0];
	matrix44[1] = (F)source44[4];
	matrix44[2] = (F)source44[8];
	matrix44[3] = (F)source44[12];
	matrix44[4] = (F)source44[1];
	matrix44[5] = (F)source44[5];
	matrix44[6] = (F)source44[9];
	matrix44[7] = (F)source44[13];
	matrix44[8] = (F)source44[2];
	matrix44[9] = (F)source44[6];
	matrix44[10] = (F)source44[10];
	matrix44[11] = (F)source44[14];
	matrix44[12] = (F)source44[3];
	matrix44[13] = (F)source44[7];
	matrix44[14] = (F)source44[11];
	matrix44[15] = (F)source44[15];

}

void RenderingOneDX11_2(LawData& data, const float* view_matrix_t, const float* projection_matrix, const double focus_distance, const RenderingProperty& rendering_property, DX11_CORE* dx11_core) {
	//各ファイルごとのレンダリング処理//

	//visual setting//
	VISUAL_SETTING vis;
	vis.atom = rendering_property.atom_draw;
	vis.atom_poly = g_visual_atompoly;
	vis.atom_radius = g_visual_atomradius;
	vis.atom_color = rendering_property.atom_color;
	vis.bond = rendering_property.bond_draw;
	vis.bond_poly = g_visual_bondpoly;
	vis.history_frame = g_visual_history_frame;
	vis.trajectory_width = g_visual_trajectory_width;

	unsigned int box_color = (unsigned int)((1.0f - g_bg_color[0])*255.f)
		| ((unsigned int)((1.0f - g_bg_color[1])*255.f) << 8)
		| ((unsigned int)((1.0f - g_bg_color[2])*255.f) << 16)
		| (0xFF000000);


	unsigned int trajectory_color = (unsigned int)((g_visual_trajectory_color[0])*255.f)
		| ((unsigned int)((g_visual_trajectory_color[1])*255.f) << 8)
		| ((unsigned int)((g_visual_trajectory_color[2])*255.f) << 16)
		| (0xFF000000);

	if (m_boxRendererDX11 == NULL) m_boxRendererDX11 = new BoxRendererDX11(dx11_core->pD3DDevice, dx11_core->pD3DDeviceContext, comShader);
	m_boxRendererDX11->SetViewAndProjectionMatrix(view_matrix_t, projection_matrix);
	state_manager_dx->StoreState();

	switch (data.type) {
	case FILETYPE_ATOM:
		{
			if (rendering_property.atom_draw == VISUAL_ATOM_POINT) {


				//レンダリング処理///////////////////////////////////////////

				if (m_pointRendererDX11 == NULL) m_pointRendererDX11 = new PointRendererDX11(dx11_core->pD3DDevice, dx11_core->pD3DDeviceContext, comShader);

				m_pointRendererDX11->SetViewAndProjectionMatrix(view_matrix_t, projection_matrix);
				ATOMS_DATA* dat = data.GetDataPointer();

				m_pointRendererDX11->Draw(dat, data.GetBond(), vis);

				if (rendering_property.box_draw) {
					mat33d axis;
					vec3d org;
					data.GetBoxSize(&axis, &org);
					m_boxRendererDX11->Draw(axis, org, 12, box_color);
				}

			} else {
				
				if (rendering_property.atom_draw == VISUAL_ATOM_SPHERE) {


					//レンダリング処理///////////////////////////////////////////
					if (vis.atom_color == VISUAL_ATOMCOLOR_HEIGHT) {

						if (m_sphere_color_table_renderer == nullptr) m_sphere_color_table_renderer = new SphereColorTableRendererDX11(dx11_core->pD3DDevice, dx11_core->pD3DDeviceContext, comShader);

						
						ATOMS_DATA* dat = data.GetDataPointer();

						static const float color_table[9 * 4] = { 0.0, 0.0, 0.5647058823529412, 1.0,  //#000090
							0.0, 0.0588235294117647, 1.0, 1.0,					//#000fff
							0.0, 0.5647058823529412, 1.0, 1.0,					//#0090ff
							0.0588235294117647, 1.0, 0.9333333333333333, 1.0,	//#0fffee
							0.5647058823529412, 1.0, 0.4392156862745098,1.0,	//#90ff70
							1.0, 0.9333333333333333, 0.0, 1.0,					//#ffee00
							1.0, 0.4392156862745098, 0.0, 1.0,					//#ff7000
							0.9333333333333333, 0.0, 0.0, 1.0,					//#ee0000
							0.4980392156862745, 0.0, 0.0, 1.0 };				//#7f0000
						
						mat33d axis;
						vec3d org;
						data.GetBoxSize(&axis, &org);

						const float height_center = g_color_hight_center > 0.0 ? g_color_hight_center : axis.c.z / 2.0;
						const float height_range = g_color_hight_range > 0.0 ? g_color_hight_range : axis.c.z / 2.0;

						m_sphere_color_table_renderer->SetViewAndProjectionMatrix(view_matrix_t, projection_matrix);
						m_sphere_color_table_renderer->Draw(dat->pcnt, dat->r, color_table, 9,
							(float)(height_center - height_range), (float)(height_center + height_range), vis);

					} else {

						if (vis.atom_color == VISUAL_ATOMCOLOR_PRESSURE) {
							MessageBox(NULL, _T("warning"), _T("color mode with pressure is not supported in this version\n"), MB_OK);
						}

						if (m_sphere_renderer == NULL) m_sphere_renderer = new SphereRendererDX11(dx11_core->pD3DDevice, dx11_core->pD3DDeviceContext, comShader);


						ATOMS_DATA* dat = data.GetDataPointer();

						const uint32_t* color_table = (vis.atom_color == VISUAL_ATOMCOLOR_BONDNUM) ? (const uint32_t*)bondnum_colors :
							(const uint32_t*)atom_colors;
						const int* color_indexes = (vis.atom_color == VISUAL_ATOMCOLOR_BONDNUM) ? (data.GetBond()->coords) :
							dat->knd;

						m_sphere_renderer->SetViewAndProjectionMatrix(view_matrix_t, projection_matrix);
						m_sphere_renderer->Draw(dat->pcnt, dat->r, color_indexes, color_table, vis);
					}
				}

				if (vis.bond) {
					if (vis.atom_color == VISUAL_ATOMCOLOR_HEIGHT) {

					} else {

						if (m_cylinder_renderer == NULL) m_cylinder_renderer = new CylinderRendererDX11(dx11_core->pD3DDevice, dx11_core->pD3DDeviceContext, comShader);


						ATOMS_DATA* dat = data.GetDataPointer();

						const uint32_t* color_table = (vis.atom_color == VISUAL_ATOMCOLOR_BONDNUM) ? (const uint32_t*)bondnum_colors :
							(const uint32_t*)atom_colors;
						const int* color_indexes = (vis.atom_color == VISUAL_ATOMCOLOR_BONDNUM) ? (data.GetBond()->coords) :
							dat->knd;

						m_cylinder_renderer->SetViewAndProjectionMatrix(view_matrix_t, projection_matrix);
						m_cylinder_renderer->Draw(data.GetBond()->count, data.GetBond()->bvector, dat->r, color_indexes, color_table, vis);
					}
				}
				

				if (rendering_property.box_draw) {
					mat33d axis;
					vec3d org;
					data.GetBoxSize(&axis, &org);
					m_boxRendererDX11->Draw(axis, org, 12, box_color);
				}

			}
		}
		break;

	case FILETYPE_FIELD:
		{
		}
		break;


	case FILETYPE_KARABA:
		{

			//レンダリング処理///////////////////////////////////////////

			if (m_karabaRendererDX11 == NULL) m_karabaRendererDX11 = new KrbRendererDX11(dx11_core->pD3DDevice, dx11_core->pD3DDeviceContext, comShader);

			m_karabaRendererDX11->SetViewAndProjectionMatrix(view_matrix_t, projection_matrix);
			KRB_INFO* dat = data.GetDataPointerKRB();

			m_karabaRendererDX11->Draw(dat, trajectory_color, vis);


			if (rendering_property.box_draw) {
				mat33d axis;
				vec3d org;
				data.GetBoxSize(&axis, &org);
				m_boxRendererDX11->Draw(axis, org, 12, box_color);
			}

		}
		break;


	case FILETYPE_TRAJ:
		{

			const uint32_t trajectory_color = (unsigned int)((g_visual_trajectory_color[0])*255.f)
				| ((unsigned int)((g_visual_trajectory_color[1])*255.f) << 8)
				| ((unsigned int)((g_visual_trajectory_color[2])*255.f) << 16)
				| (0xFF000000);


			const TrajectoryData* traj = data.GetTrajectory();
			if (m_strip_renderer == NULL) m_strip_renderer = new CylinderStripRendererDX11(dx11_core->pD3DDevice, dx11_core->pD3DDeviceContext, comShader);
			m_strip_renderer->SetViewAndProjectionMatrix(view_matrix_t, projection_matrix);
			m_strip_renderer->Draw(traj->positions, traj->strip_indexes, traj->strip_lengths, traj->num_strips, trajectory_color, vis);

			if (vis.atom == VISUAL_ATOM_SPHERE) {
				
				vector<vec3f> positions;
				vector<int> color_indexes;
				int k = 0;
				for (int i = 0; i < traj->num_strips; ++i) {
					positions.push_back(traj->positions[traj->strip_indexes[k]]);
					k += traj->strip_lengths[i];
					positions.push_back(traj->positions[traj->strip_indexes[k-1]]);
					
					color_indexes.push_back(0);
					color_indexes.push_back(1);
				}
				
				if (m_sphere_renderer == NULL) m_sphere_renderer = new SphereRendererDX11(dx11_core->pD3DDevice, dx11_core->pD3DDeviceContext, comShader);
				m_sphere_renderer->SetViewAndProjectionMatrix(view_matrix_t, projection_matrix);

				if (positions.empty()) {
					int ci[1] = { 0 };
					m_sphere_renderer->Draw(1, traj->positions, ci, (const uint32_t*)atom_colors, vis);
				} else {
					m_sphere_renderer->Draw(positions.size(), &(positions[0]), &(color_indexes[0]), (const uint32_t*)atom_colors, vis);
				}
			}

			if (rendering_property.box_draw) {
				mat33d axis;
				vec3d org;
				data.GetBoxSize(&axis, &org);
				m_boxRendererDX11->Draw(axis, org, 12, box_color);
			}
		}
		break;

	case FILETYPE_PMD:
		{

			if (m_vntRenderer == NULL) {
				m_vntRenderer = new VNTRenderer(dx11_core->pD3DDevice, dx11_core->pD3DDeviceContext, comShader);
			}

			m_vntRenderer->SetViewAndProjectionMatrix(view_matrix_t, projection_matrix);


			MATERIAL_INFO* materials;
			IPolyRepository* pmd = data.GetPolyRepository();
			int num_materials = pmd->GetMaterialList(&materials);
			UINT vertex_stride;
			UINT index_stride;
			LPVOID vertex_buffer;
			int vertex_count = pmd->GetVertexBuffer(&vertex_buffer, &vertex_stride);
			LPVOID index_buffer;
			int index_count = pmd->GetIndexBuffer(&index_buffer, &index_stride);


			vec3f r = { 0.0f, 0.0f, 0.0f };
			if (num_materials > 0) {
				m_vntRenderer->UpdateVertexBuffer(vertex_buffer, vertex_count, vertex_stride);//頂点データをVRAMに転送
				m_vntRenderer->UpdateIndexBuffer(index_buffer, index_count, index_stride);	//インデックスデータをVRAMに転送
				m_vntRenderer->DrawMaterials(num_materials, materials, &r);
			}



		}
		break;
	case FILETYPE_VMEC:
		{


			//レンダリング処理///////////////////////////////////////////

			if (m_pointSpriteDX11 == NULL) {
				m_pointSpriteDX11 = new PointSpriteDX11(dx11_core->pD3DDevice, dx11_core->pD3DDeviceContext, comShader);
			}

			m_pointSpriteDX11->SetViewAndProjectionMatrix(view_matrix_t, projection_matrix);
			VMEC_Repository* vmec = data.GetVmec();
			float* values = vmec->GetFieldPointer();
			size_t count = vmec->GetNumParticles();
			float range_max[3], range_min[3];
			float alpha;
			vmec->GetEffectiveRange(range_max, range_min, &alpha);
			int stride = vmec->GetStride();

			mat33d boxaxis;
			vec3d boxorg;
			vmec->GetBoxaxis(&boxaxis, &boxorg);
			//double width = 0.001 * sqrt((boxaxis.a)*(boxaxis.a) + (boxaxis.b)*(boxaxis.b) + (boxaxis.c)*(boxaxis.c));
			double width = 0.001 * sqrt((boxaxis.a)*(boxaxis.a) + (boxaxis.b)*(boxaxis.b) + (boxaxis.c)*(boxaxis.c));

			m_pointSpriteDX11->SetPointSize((float)width);

			m_pointSpriteDX11->Draw(values, count, stride, range_max, range_min, alpha);

		}
		break;

	}
	state_manager_dx->RestoreState();


}


void Draw2D(DX11_CORE* dx11_core) {

	if (m_vntRenderer == NULL) {
		m_vntRenderer = new VNTRenderer(dx11_core->pD3DDevice, dx11_core->pD3DDeviceContext, comShader);
	}
	
	mat44f mat_identify;
	mat_identify.SetIdentity();
	m_vntRenderer->SetViewAndProjectionMatrix((float*)&mat_identify, (float*)&mat_identify);

	//WCHAR test_texture[] = _T("C:\\physics\\VisualStudio\\AIScope3\\対応ファイル仕様書\\pmd\\ion-boy_facetxt.bmp");
	WCHAR test_texture[] = _T("C:\\physics\\VisualStudio\\AIScope3\\AIScope3_vs2015_git\\resource\\button0.png");
	//MATERIAL_INFO* materials;
	MATERIAL_INFO materials[1];
	materials[0].indexCount = 6;
	materials[0].indexOffset = 0;
	materials[0].textureid = -1;
	materials[0].texture_name = test_texture;
	//materials[0].texture_name = NULL;
	materials[0].diffuse_color[0] = 0.5f;
	materials[0].diffuse_color[1] = 0.5f;
	materials[0].diffuse_color[2] = 0.5f;
	materials[0].diffuse_color[3] = 1.0f;
	//IPolyRepository* pmd = data.GetPolyRepository();
	int num_materials = 1;// pmd->GetMaterialList(&materials);
	UINT vertex_stride = sizeof(float) * 8;
	UINT index_stride = 4;

	struct RectF {
		float x0;
		float y0;
		float x1;
		float y1;
	}; 

	RectF button_position = {-1.0f,-1.0f, -0.9f, -0.9f};
	RectF button_coords = { 0.0f,0.0f, 1.0f, 1.0f };

	VERTEX_F3F3F2 vertexes[4];
	vertexes[2].position.Set(button_position.x1, button_position.y0, 0.0);
	vertexes[2].normal.Set(0.0, 0.0, -1.0);
	vertexes[2].coord_u = button_coords.x1;
	vertexes[2].coord_v = button_coords.y1;
	
	vertexes[3].position.Set(button_position.x1, button_position.y1, 0.0);
	vertexes[3].normal.Set(0.0, 0.0, -1.0);
	vertexes[3].coord_u = button_coords.x1;
	vertexes[3].coord_v = button_coords.y0;

	vertexes[1].position.Set(button_position.x0, button_position.y1,0.0);
	vertexes[1].normal.Set(0.0, 0.0, -1.0);
	vertexes[1].coord_u = button_coords.x0;
	vertexes[1].coord_v = button_coords.y0;

	vertexes[0].position.Set(button_position.x0, button_position.y0, 0.0);
	vertexes[0].normal.Set(0.0, 0.0, -1.0);
	vertexes[0].coord_u = button_coords.x0;
	vertexes[0].coord_v = button_coords.y1;

	LPVOID vertex_buffer = (LPVOID)vertexes;
	int vertex_count = 4;// pmd->GetVertexBuffer(&vertex_buffer, &vertex_stride);
	//LPVOID index_buffer;
	int index_buffer[6] = { 0,1,2, 1,2,3 };
	int index_count = 6;// pmd->GetIndexBuffer(&index_buffer, &index_stride);


	vec3f r = { 0.0f, 0.0f, 0.0f };
	if (num_materials > 0) {
		m_vntRenderer->UpdateVertexBuffer(vertex_buffer, vertex_count, vertex_stride);//頂点データをVRAMに転送
		m_vntRenderer->UpdateIndexBuffer(index_buffer, index_count, index_stride);	//インデックスデータをVRAMに転送
		m_vntRenderer->DrawMaterials(num_materials, materials, &r);
	}



}

void ClearRenderingTargetDX11() {
	DX11_CORE* dx11_core = &s_dx11_core;

	//BackBufferのクリア
	dx11_core->pD3DDeviceContext->ClearRenderTargetView(dx11_core->pRenderTargetView, g_bg_color);

	//DepthBufferのクリア
	dx11_core->pD3DDeviceContext->ClearDepthStencilView(dx11_core->pDepthStencilView, D3D11_CLEAR_DEPTH, 1.0f, 0);

}

void RenderingCoreDX11(vector<LawData>& data_list,  int screen_w, int screen_h, int range_x, int range_y, int range_w, int range_h, const double* view_matrix, const double focus_distance, const RenderingProperty& rendering_property)
{
	DX11_CORE* dx11_core = &s_dx11_core;

	
	//変換行列計算
	float view_matrix_t[16];
	Transpose<float, double>(view_matrix_t, view_matrix);

	mat44f projection_matrix;

	const int num_files =  data_list.size();
	if ((rendering_property.multiview_mode == 0) || (num_files == 1)) {
		// Setup the viewport
		D3D11_VIEWPORT vp;
		vp.Width = (FLOAT)range_w;
		vp.Height = (FLOAT)range_h;
		vp.MinDepth = 0.0f;
		vp.MaxDepth = 1.0f;
		vp.TopLeftX = 0;
		vp.TopLeftY = 0;
		dx11_core->pD3DDeviceContext->RSSetViewports(1, &vp);

		const int image_w = screen_w;
		const int image_h = screen_h;

		SetProjectionRange(rendering_property.projection_mode, image_w, image_h, range_x, range_y, range_w, range_h, focus_distance, &projection_matrix);




		for (auto it = data_list.begin(), it_end = data_list.end(); it != it_end; ++it) {
			
			RenderingOneDX11_2(*it, view_matrix_t, (float*)&projection_matrix, focus_distance, rendering_property, dx11_core);
		}
	} else {


		//スクリーンサイズ比からイメージを並べる数を計算//
		double dnx = sqrt((double)num_files * (double)screen_w / (double)screen_h);
		int multiview_nx = (int)(dnx)+1;
		if (multiview_nx > num_files) { multiview_nx = num_files; }
		int multiview_ny = num_files / multiview_nx;
		if (num_files > multiview_nx * multiview_ny) { multiview_ny++; }


		int nx = 0;
		int ny = 0;// multiview_ny - 1;
		for (auto it = data_list.begin(), it_end = data_list.end(); it != it_end; ++it) {

			

			//一枚当たりのイメージサイズ//
			const int image_x = (screen_w * nx) / multiview_nx;
			const int image_w = (screen_w * (nx + 1)) / multiview_nx - image_x;
			const int image_y = (screen_h * ny) / multiview_ny;
			const int image_h = (screen_h * (ny + 1)) / multiview_ny - image_y;

			//要求描画範囲とオーバーラップがあるか確認//
			if (range_x + range_w > image_x) {
				if (image_x + image_w > range_x) {
					if (range_y + range_h > image_y) {
						if (image_y + image_h > range_y) {

							//オーバーラップ部分の範囲(ただしimage_x,yからの相対位置)
							const int real_range_x = max(range_x - image_x, 0);
							const int real_range_w = min(range_x + range_w - image_x, image_w) - real_range_x;
							const int real_range_y = max(range_y - image_y, 0);
							const int real_range_h = min(range_y + range_h - image_y, image_h) - real_range_y;

							//オーバーラップ部分の範囲(ただしrange_x,yからの相対位置)
							const int view_x = max(0, image_x - range_x);
							const int view_w = min(range_w, image_x + image_w - range_x) - view_x;
							const int view_y = max(0, image_y - range_y);
							const int view_h = min(range_h, image_y + image_h - range_y) - view_y;

							//glViewport(view_x, view_y, view_w, view_h);
							D3D11_VIEWPORT vp;
							vp.Width = (FLOAT)view_w;
							vp.Height = (FLOAT)view_h;
							vp.MinDepth = 0.0f;
							vp.MaxDepth = 1.0f;
							vp.TopLeftX = view_x;
							vp.TopLeftY = view_y;
							dx11_core->pD3DDeviceContext->RSSetViewports(1, &vp);


							SetProjectionRange(rendering_property.projection_mode, image_w, image_h, real_range_x, real_range_y, real_range_w, real_range_h, focus_distance, &projection_matrix);
							//SetProjectionGL( image_w, image_h, cam->GetForcusDistance() );
							//RenderingOneGL15(fi, view_matrix, focus_distance, rendering_property);
							RenderingOneDX11_2(*it, view_matrix_t, (float*)&projection_matrix, focus_distance, rendering_property, dx11_core);

						}
					}
				}
			}
			nx++;
			if (nx == multiview_nx) {
				nx = 0;
				//ny--;
				ny++;
			}


		}

	}

	/*
	操作ボタンの表示
	Draw2D(dx11_core);
	*/
}

void CheckRtvDX11() {
#ifdef _DEBUG
	{
		ID3D11RenderTargetView* org_RTV;
		ID3D11DepthStencilView* org_DSV;
		s_dx11_core.pD3DDeviceContext->OMGetRenderTargets(1, &org_RTV, &org_DSV);
		SAFE_RELEASE(org_RTV);
		SAFE_RELEASE(org_DSV);
	}
#endif
}

void RenderingDX11(vector<LawData>& data_list, int screen_w, int screen_h, int range_x, int range_y, int range_w, int range_h, const double* view_matrix, const double focus_distance, const RenderingProperty& rendering_property) {
	
	
	switch ((g_filter_mode == 1) ? FILTER_DEPTH_OF_FIELD : FILTER_NONE) {
	case FILTER_NONE:
		ClearRenderingTargetDX11();
		RenderingCoreDX11(data_list, screen_w, screen_h, range_x, range_y, range_w, range_h, view_matrix, focus_distance, rendering_property);
		break;
	case FILTER_DEPTH_OF_FIELD:
		s_filterDoF->Render(data_list, screen_w, screen_h, range_x, range_y, range_w, range_h, view_matrix, focus_distance, rendering_property);
		break;
	}
	
}


///////////////////////////////////////////////////////////////////
//変換行列用ライブラリ
///////////////////////////////////////////////////////////////////

namespace {
	void SetProjectionPerspectiveMatrix(float angle, float wph, float z_near, float z_far, mat44f* pProjectionMatrix) {
		pProjectionMatrix->Clear();

		pProjectionMatrix->m11 = 1.0f / tan(angle);
		pProjectionMatrix->m22 = pProjectionMatrix->m11 * wph;
		pProjectionMatrix->m33 = z_far / (z_far - z_near);
		pProjectionMatrix->m43 = -z_near * pProjectionMatrix->m33;
		pProjectionMatrix->m34 = 1.0f;
		pProjectionMatrix->m44 = 0.0f;


	}

	void SetProjectionMatrixFrustnum(double left, double right, double bottom, double top,
		double zNear, double zFar, mat44f* pProjectionMatrix) {

		pProjectionMatrix->Clear();

		const double width = right - left;
		const double height = top - bottom;
		const double zwidth = zFar - zNear;

		pProjectionMatrix->m11 = (float)(2.0*zNear / width);
		pProjectionMatrix->m31 = -(float)((right + left) / width);
		pProjectionMatrix->m22 = (float)(2.0*zNear / height);
		pProjectionMatrix->m32 = -(float)((top + bottom) / height);
		//pProjectionMatrix->m33 = (float)((zFar + zNear) / zwidth);
		//pProjectionMatrix->m43 = -(float)((zFar*zNear) / zwidth);
		pProjectionMatrix->m33 = (float)((zFar) / zwidth);
		pProjectionMatrix->m43 = -(float)((zFar*zNear) / zwidth);
		pProjectionMatrix->m34 = 1.0f;
		pProjectionMatrix->m44 = 0.0f;

		/* OpenGL(右手座標系)でのPorjectionMaxtixの定義に加えて,
		 z座標のみ反転する次のmatrixを掛けたもの

		  ( 1  0  0  0
			0  1  0  0
			0  0 -1  0
			0  0  0  1  )
		 */



	}


	void SetProjectionMatrixOrtho(double left, double right, double bottom, double top,
		double zNear, double zFar, mat44f* pProjectionMatrix) {

		pProjectionMatrix->Clear();

		const double width = right - left;
		const double height = top - bottom;
		const double zwidth = zFar - zNear;

		pProjectionMatrix->m11 = (float)(2.0 / width);
		pProjectionMatrix->m41 = -(float)((right + left) / width);
		pProjectionMatrix->m22 = (float)(2.0 / height);
		pProjectionMatrix->m42 = -(float)((top + bottom) / height);
		
		pProjectionMatrix->m33 = (float)(1.0 / zwidth);
		pProjectionMatrix->m43 = -(float)((zNear) / zwidth);
		pProjectionMatrix->m44 = 1.0f;
		
		
		
		/* OpenGL(右手座標系)でのPorjectionMaxtixの定義に加えて,
		z座標のみ反転する次のmatrixを掛けたもの

		( 1  0  0  0
		0  1  0  0
		0  0 -1  0
		0  0  0  1  )
		*/
	}




	//一つのデータの可視化に関して、描画範囲をPrjection範囲として設定//
	void SetProjectionRange(int projection_mode, int image_w, int image_h, int range_x, int range_y, int range_w, int range_h, double distance, mat44f* projection_matrix) {

		//射影変換の設定===================================//
		if (projection_mode == 1) {  //正射影//
			double top = distance * tan(VIEW_ANGLE);
			double right = top * (double)image_w / (double)image_h;

			double dh = top * 2.0 * ((double)range_h / (double)image_h);
			double dw = right * 2.0 * ((double)range_w / (double)image_w);

			double btm = top * (-1.0 + 2.0 * ((double)range_y / (double)image_h));
			double left = right * (-1.0 + 2.0 * ((double)range_x / (double)image_w));

			//SetProjectionMatrixOrtho(left, left + dw, btm, btm + dh, VIEW_NEAR*distance, VIEW_FAR * distance, projection_matrix);
			double top2 = top * (1.0 - 2.0 * ((double)range_y / (double)image_h));
			SetProjectionMatrixOrtho(left, left + dw, top2 - dh, top2, VIEW_NEAR*distance, VIEW_FAR * distance/5.0, projection_matrix);

		} else {
			double top = VIEW_NEAR * distance * tan(VIEW_ANGLE);
			double right = top * (double)image_w / (double)image_h;

			double dh = top * 2.0 * ((double)range_h / (double)image_h);
			double dw = right * 2.0 * ((double)range_w / (double)image_w);

			double btm = top * (-1.0 + 2.0 * ((double)range_y / (double)image_h));
			double left = right * (-1.0 + 2.0 * ((double)range_x / (double)image_w));


			//SetProjectionMatrixFrustnum(left, left + dw, btm, btm + dh, VIEW_NEAR*distance, VIEW_FAR * distance, projection_matrix);
			double top2 = top * (1.0 - 2.0 * ((double)range_y / (double)image_h));
			SetProjectionMatrixFrustnum(left, left + dw, top2 - dh, top2, VIEW_NEAR*distance, VIEW_FAR * distance, projection_matrix);
		}
		//===================================射影変換//
	}
}

#if 0
//DirectXで描画した内容をビットマップ画像として保存//
void OutputDX11(vector<LawData>& data_list, const TCHAR *fName, int width, int height, int magnify, const double* view_matrix, const double focus_distance, const RenderingProperty& rendering_property){
	
	DX11_CORE* dx11_core = &s_dx11_core;

	FILE* fp = _tfopen(fName,_T("wb"));
	if (!fp){
		printf("bmpファイルの作成に失敗しました\n");
		return;
	}
	
	WriteBMP wbmp(width * magnify, height * magnify, 3);


	
	
    //(2)スクリーンフォーマットを取得//
    ID3D11Texture2D* pBackBuffer = NULL;
    HRESULT hr = dx11_core->pSwapChain->GetBuffer( 0, __uuidof( ID3D11Texture2D ), ( LPVOID* )&pBackBuffer );
    if( FAILED( hr ) ) return;
	
	D3D11_TEXTURE2D_DESC descBackBuffer;
	pBackBuffer->GetDesc(&descBackBuffer);
	pBackBuffer->Release();

	
	//(3)CPU読み出し可能なレンダリングターゲット(texture)バッファをGPU上に作成//
    D3D11_TEXTURE2D_DESC Texture2DDesc;
    Texture2DDesc.ArraySize = descBackBuffer.ArraySize;
    Texture2DDesc.BindFlags = D3D11_BIND_RENDER_TARGET;
    Texture2DDesc.CPUAccessFlags = D3D11_CPU_ACCESS_READ;  //CPUで読めるようにする
    Texture2DDesc.Format = descBackBuffer.Format;
    Texture2DDesc.Height = descBackBuffer.Height;
    Texture2DDesc.Width = descBackBuffer.Width;
    Texture2DDesc.MipLevels = descBackBuffer.MipLevels;
    Texture2DDesc.MiscFlags = descBackBuffer.MiscFlags;
    Texture2DDesc.SampleDesc.Count = descBackBuffer.SampleDesc.Count;
    Texture2DDesc.SampleDesc.Quality = descBackBuffer.SampleDesc.Quality;
    Texture2DDesc.Usage = D3D11_USAGE_STAGING;               //GPUで書き込むにはD3D11_USAGE_DEFAULTにする必要がある//

    ID3D11Texture2D *hCaptureTexture;
    hr = dx11_core->pD3DDevice->CreateTexture2D( &Texture2DDesc, 0, &hCaptureTexture );//失敗する//

	//(3.5)レンダリングターゲットの差し替え//
	ID3D11RenderTargetView* renderTargetView;
	hr = dx11_core->pD3DDevice->CreateRenderTargetView(hCaptureTexture, NULL, &renderTargetView);
	ID3D11RenderTargetView* renderTV_org;
	ID3D11DepthStencilView* depthSV_org;
	dx11_core->pD3DDeviceContext->OMGetRenderTargets(1, &renderTV_org, &depthSV_org);
	dx11_core->pD3DDeviceContext->OMSetRenderTargets(1, &renderTargetView, depthSV_org);

	
	
	BYTE *buffer = NULL;




	//int offset_y = 0;

	for (int y = 0; y < magnify; y++) {
		//int offset_x = 0;
		for (int x = 0; x < magnify; x++) {


			//(1)DirectX11でレンダリングする関数.フリップはしない//
			//Clear RenderingTarget and DepthBuffer//
			dx11_core->pD3DDeviceContext->ClearRenderTargetView(renderTargetView, g_bg_color);			
			dx11_core->pD3DDeviceContext->ClearDepthStencilView(depthSV_org, D3D11_CLEAR_DEPTH, 1.0f, 0);			
			RenderingDX11_2(data_list, width*magnify, height*magnify, width*x, height*y, width, height, view_matrix, focus_distance, rendering_property);


			/*
			//(4)作成したCPU読み込み可能バッファにGPU上でデータをコピー//
			ID3D11Resource *hResource;
			dx11_core->pRenderTargetView->GetResource(&hResource);
			dx11_core->pD3DDeviceContext->CopyResource(hCaptureTexture, hResource);
			hResource->Release();
			//SaveWICTextureToFile
			*/
			//(5)GPU上の読み込み可能バッファのメモリアドレスのマップを開く//
			D3D11_MAPPED_SUBRESOURCE mappedResource;
			dx11_core->pD3DDeviceContext->Map(hCaptureTexture, 0, D3D11_MAP_READ, 0, &mappedResource);

			//(6)CPU上のメモリにバッファを確保//
			int src_stride = mappedResource.RowPitch;
			if (buffer == NULL) {
				buffer = new BYTE[src_stride * height];
			}

			//(7)GPU上の読み込み可能バッファからCPU上のバッファへ転送, マップを閉じる//
			CopyMemory(buffer, mappedResource.pData, src_stride * height);
			dx11_core->pD3DDeviceContext->Unmap(hCaptureTexture, 0);
			

			//(8)CPU上のバッファの画像をファイル書き出し//
			
			
			if (descBackBuffer.Format == DXGI_FORMAT_R16G16B16A16_FLOAT) {
				
				wbmp.CreateDIBfromRGBA_FP16(buffer, src_stride, width * x, height * y, width, height);
			} else if (descBackBuffer.Format == DXGI_FORMAT_R8G8B8A8_UNORM) {
				/*
				BITMAPINFOHEADER bmih;
				dib_buf = CreateDIBfromRGBA(buffer, src_stride, width * magnify, height * magnify, &bmih);
				writeDIB(fp, &bmih, dib_buf);
				*/
			}
			
		

		}
		

	}

	//(9)RenderTargetをデフォルトに戻す//	
	dx11_core->pD3DDeviceContext->OMSetRenderTargets(1, &renderTV_org, depthSV_org);
	renderTargetView->Release();
	hCaptureTexture->Release();

	wbmp.WriteDIB(fp);

	delete[] buffer;
	fclose(fp);



}

#else
//DirectXで描画した内容をビットマップ画像として保存//
void OutputDX11(vector<LawData>& data_list, const TCHAR *fName, int width, int height, int magnify, const double* view_matrix, const double focus_distance, const RenderingProperty& rendering_property) {

	DX11_CORE* dx11_core = &s_dx11_core;

	FILE* fp = _tfopen(fName, _T("wb"));
	if (!fp) {
		printf("bmpファイルの作成に失敗しました\n");
		return;
	}

	WriteBMP wbmp(width * magnify, height * magnify, 3);




	//(2)スクリーンフォーマットを取得//
	ID3D11Texture2D* pBackBuffer = NULL;
	HRESULT hr = dx11_core->pSwapChain->GetBuffer(0, __uuidof(ID3D11Texture2D), (LPVOID*)&pBackBuffer);
	if (FAILED(hr)) return;

	D3D11_TEXTURE2D_DESC descBackBuffer;
	pBackBuffer->GetDesc(&descBackBuffer);
	pBackBuffer->Release();


	//(3)CPU読み出し可能なバッファをGPU上に作成//
	D3D11_TEXTURE2D_DESC Texture2DDesc;
	Texture2DDesc.ArraySize = 1;
	Texture2DDesc.BindFlags = 0;
	Texture2DDesc.CPUAccessFlags = D3D11_CPU_ACCESS_READ;
	Texture2DDesc.Format = descBackBuffer.Format;
	Texture2DDesc.Height = descBackBuffer.Height;
	Texture2DDesc.Width = descBackBuffer.Width;
	Texture2DDesc.MipLevels = 1;
	Texture2DDesc.MiscFlags = 0;
	Texture2DDesc.SampleDesc.Count = 1;
	Texture2DDesc.SampleDesc.Quality = 0;
	Texture2DDesc.Usage = D3D11_USAGE_STAGING;

	ID3D11Texture2D *hCaptureTexture;
	dx11_core->pD3DDevice->CreateTexture2D(&Texture2DDesc, 0, &hCaptureTexture);

	BYTE *buffer = NULL;

	//int offset_y = 0;

	for (int y = 0; y < magnify; y++) {
		//int offset_x = 0;
		for (int x = 0; x < magnify; x++) {


			//(1)DirectX11でレンダリングする関数.フリップはしない//
			//ClearRenderingTargetDX11();
			RenderingDX11(data_list, width*magnify, height*magnify, width*x, height*y, width, height, view_matrix, focus_distance, rendering_property);



			//(4)作成したCPU読み込み可能バッファにGPU上でデータをコピー//
			ID3D11Resource *hResource;
			dx11_core->pRenderTargetView->GetResource(&hResource);
			dx11_core->pD3DDeviceContext->CopyResource(hCaptureTexture, hResource);
			hResource->Release();
			//SaveWICTextureToFile

			//(5)GPU上の読み込み可能バッファのメモリアドレスのマップを開く//
			D3D11_MAPPED_SUBRESOURCE mappedResource;
			dx11_core->pD3DDeviceContext->Map(hCaptureTexture, 0, D3D11_MAP_READ, 0, &mappedResource);

			//(6)CPU上のメモリにバッファを確保//
			int src_stride = mappedResource.RowPitch;
			if (buffer == NULL) {
				buffer = new BYTE[src_stride * height];
			}

			//(7)GPU上の読み込み可能バッファからCPU上のバッファへ転送, マップを閉じる//
			CopyMemory(buffer, mappedResource.pData, src_stride * height);
			dx11_core->pD3DDeviceContext->Unmap(hCaptureTexture, 0);


			//(8)CPU上のバッファの画像をファイル書き出し//


			if (descBackBuffer.Format == DXGI_FORMAT_R16G16B16A16_FLOAT) {

				wbmp.CreateDIBfromRGBA_FP16(buffer, src_stride, width * x, height * y, width, height);
			} else if (descBackBuffer.Format == DXGI_FORMAT_R8G8B8A8_UNORM) {
				/*
				BITMAPINFOHEADER bmih;
				dib_buf = CreateDIBfromRGBA(buffer, src_stride, width * magnify, height * magnify, &bmih);
				writeDIB(fp, &bmih, dib_buf);
				*/
			}



		}


	}
	hCaptureTexture->Release();

	wbmp.WriteDIB(fp);

	delete[] buffer;
	fclose(fp);



}
#endif
