//=====================================
// 複数インスタンス表示クラス
//
//
// 
//=====================================
#pragma once



#include <d3d11.h>
#include <vector>

#include "trajectory2.h"	//ATOMS_DATAのためだけに必要//

#include "MDLoader.h"	//ATOMS_DATAのためだけに必要//
#include "BOND_INFO.h"

//#include "sdkwgl.h"
#include "vec3.h"
#include "mat33.h"
#include "mat44.h"
#include "io64.h"
#include "mod.h"


#include "visual.h"


#include "BasicRenderer.h"
#include "VertexTypes.h"
#include "GeometrySphere.h"
#include "GeometryCylinder.h"

#include "aidx_constantbuffer.h"


class SphereColorTableRendererDX11 {
private:
	
	RendererWrapper renderer;
		
	int m_vram_size;
	int m_vbo_vcount;
	VERTEX_F3F3* m_vbo_vertex;
	int m_vbo_icount;
	int* m_vbo_indexes;
	unsigned int m_index_type;
	std::vector<VERTEX_F4> m_instances;
	
	//ポリゴンモデル用のバッファ
	GeometrySphere* m_geometry_sphere;

	VISUAL_SETTING m_prev_vis;
		
	AIDX11::ConstantBuffer m_CB_color_table;
	AIDX11::ConstantBuffer m_CB_range;

	void InitShader();
	void CreateSphere();
public:
	SphereColorTableRendererDX11(ID3D11Device* pD3DDevice, ID3D11DeviceContext* pD3DDeviceContext, CommonShader* comShader);
	virtual ~SphereColorTableRendererDX11();



	void Draw(int num_particles, const vec3f* r, const float* color_table, const int num_colors, float height_min, float height_max, const VISUAL_SETTING& vis);

	void SetViewAndProjectionMatrix(const float* pViewMatrix, const float* pProjectionMatrix) {
		renderer.SetViewAndProjectionMatrix(pViewMatrix, pProjectionMatrix);
	}

};

/*
struct PairIndex {
	int i;
	int j;
};
*/

class CylinderColorTableRendererDX11 {
private:

	RendererWrapper renderer;

	
	int m_vram_size;
	int m_vbo_vcount;
	VERTEX_F3F3UI1* m_vbo_vertex;
	int m_vbo_icount;
	int* m_vbo_indexes;
	unsigned int m_index_type;


	//void CreateCylinder(int start, int count, const int* pair_indexes, const vec3f* r, const int* color_indexes, const uint32_t* color_table);
	void CreateBond(int start, int count, const VECTOR_IJ* bonds, const vec3f* r, const int* color_indexes, const uint32_t* color_table);

	//ポリゴンモデル用のバッファ
	GeometryCylinder* m_geometry_bond;

	VISUAL_SETTING m_prev_vis;

	int m_visual_bond;	//ボンドの表示種類
	
	void InitShader();

public:
	CylinderColorTableRendererDX11(ID3D11Device* pD3DDevice, ID3D11DeviceContext* pD3DDeviceContext, CommonShader* comShader);
	virtual ~CylinderColorTableRendererDX11();



	void Draw(int num_pairs, const VECTOR_IJ* bonds, const vec3f* r, const int* color_indexes, const uint32_t* color_table, const VISUAL_SETTING& vis);

	void SetViewAndProjectionMatrix(const float* pViewMatrix, const float* pProjectionMatrix) {
		renderer.SetViewAndProjectionMatrix(pViewMatrix, pProjectionMatrix);
	}

};




