
#ifndef __lmc_reader_h__
#define __lmc_reader_h__

#pragma once


#include <stdio.h>
typedef char LMC_INT;


struct LMC_EVENT{
	int grid_index;
	LMC_INT delta;
};

class LMC_Reader
{
private:


	struct ATOMIC_ELEMENT{
		char symbol[4];		//need to include '\0'//
	};

	FILE* m_fp;	//ファイル
	//int m_state;	//0:ヘッダー読み込み前,1:ヘッダー読み込み済み,2:ファイル末尾//
	int m_grid_x;
	int m_grid_y;
	int m_grid_z;
	double m_boxaxis[12];
	ATOMIC_ELEMENT m_elements[108];
	int m_num_elements;

	int m_state;
	double m_next_step_time;

	int ReadHeader();
	int GetDataIndex(const char* atomicsymbol, const ATOMIC_ELEMENT* elements, int num_elements);

public:


	LMC_Reader();
	virtual ~LMC_Reader();

	///////////////////////////////////////////////////////
	//ファイルをオープンする.
	//
	int Open(const char* filepath);
	//
	//	正しくオープンできたときは 0 を返す.
	//	既に別のファイルをオープン済みの場合は -2 を返す.
	//	指定したファイルが見つからないときは -1 を返す.
	//	読み込んだデータが壊れているとき -3を返す.
	//	
	/////////////////////////////////////////////////////////////


	///////////////////////////////////////////////////////
	//ファイルを閉じる.
	//
	void Close();
	//
	//	delete してもこの関数が自動で呼ばれる.
	//
	///////////////////////////////////////////////////////////

	////////////////////////////////////////////////////////////////////////////////
	//ファイルを読み込みして必要なデータを返す.
	//
	void GetNumGrid(int* grid_x, int* grid_y, int* grid_z);
	//
	//	粒子数を返す.
	//	Readの後にこの関数を呼ぶと次のフレームに進める.
	//
	//
	///////////////////////////////////////////////////////////
	
	////////////////////////////////////////////////////////////////////////////////
	//ファイルを読み込みして必要なデータを返す.
	//
	int GetNumElements();
	//
	//	粒子数を返す.
	//	Readの後にこの関数を呼ぶと次のフレームに進める.
	//
	//
	///////////////////////////////////////////////////////////

	
	////////////////////////////////////////////////////////////////////////////////
	//ファイルを読み込みして必要なデータを返す.
	//
	void GetBoxSize(double* boxaxis);
	//
	//
	///////////////////////////////////////////////////////////

	////////////////////////////////////////////////////////////////////////////////
	//初期状態を読み込む
	//
	int ReadInitialGrid(LMC_INT* buffer);
	//
	//
	//
	////////////////////////////////////////////////////////////////////////////////

	////////////////////////////////////////////////////////////////////////////////
	//1ステップの読み込み//
	//
	int ReadStep(double* time, LMC_EVENT* events);
	//
	//  戻り値は1ステップのイベント数
	//  -1が帰ってきたらファイル終端で終了
	//
	//
	///////////////////////////////////////////////////////////////////////////////
	

};






#endif	//!__lmc_reader_h__
